@@ -83,7 +83,7 @@
  * specific kinds of stats.
  *
  *
- * Copyright (c) 2001-2023, PostgreSQL Global Development Group
+ * Copyright (c) 2001-2024, PostgreSQL Global Development Group
  *
  * IDENTIFICATION
  *	  src/backend/utils/activity/pgstat.c
@@ -93,7 +93,6 @@
 
 #include <unistd.h>
 
-#include "access/transam.h"
 #include "access/xact.h"
 #include "lib/dshash.h"
 #include "pgstat.h"
@@ -101,8 +100,6 @@
 #include "storage/fd.h"
 #include "storage/ipc.h"
 #include "storage/lwlock.h"
-#include "storage/pg_shmem.h"
-#include "storage/shmem.h"
 #include "utils/guc_hooks.h"
 #include "utils/memutils.h"
 #include "utils/pgstat_internal.h"
@@ -603,6 +600,7 @@ pgstat_report_stat(bool force)
 		!have_slrustats &&
 		!pgstat_have_pending_wal())
 	{
+		Assert(pending_since == 0);
 		return 0;
 	}
 
@@ -824,9 +822,6 @@ pgstat_fetch_entry(PgStat_Kind kind, Oid dboid, Oid objoid)
 
 	pgstat_prep_snapshot();
 
-	/* clear padding */
-	memset(&key, 0, sizeof(struct PgStat_HashKey));
-
 	key.kind = kind;
 	key.dboid = dboid;
 	key.objoid = objoid;
@@ -1398,15 +1393,7 @@ pgstat_write_statsfile(void)
 
 		CHECK_FOR_INTERRUPTS();
 
-		/*
-		 * We should not see any "dropped" entries when writing the stats
-		 * file, as all backends and auxiliary processes should have cleaned
-		 * up their references before they terminated.
-		 *
-		 * However, since we are already shutting down, it is not worth
-		 * crashing the server over any potential cleanup issues, so we simply
-		 * skip such entries if encountered.
-		 */
+		/* we may have some "dropped" entries not yet removed, skip them */
 		Assert(!ps->dropped);
 		if (ps->dropped)
 			continue;
@@ -1652,16 +1639,6 @@ pgstat_read_statsfile(void)
 
 					header = pgstat_init_entry(key.kind, p);
 					dshash_release_lock(pgStatLocal.shared_hash, p);
-					if (header == NULL)
-					{
-						/*
-						 * It would be tempting to switch this ERROR to a
-						 * WARNING, but it would mean that all the statistics
-						 * are discarded when the environment fails on OOM.
-						 */
-						elog(ERROR,	"could not allocate entry %d/%u/%u",
-							 key.kind, key.dboid, key.objoid);
-					}
 
 					if (!read_chunk(fpin,
 									pgstat_get_entry_data(key.kind, header),
